#ifndef SLMATH_FLOAT_UTIL_H
#define SLMATH_FLOAT_UTIL_H

#include <slmath/vec_templ_func.h>

SLMATH_BEGIN()

/**
 * \defgroup float_util Floating point helper functions.
 * @ingroup slmath
 */
/*@{*/

/**
 * Returns cotangent of z.
 * @param z tan(z) must be != 0
 * @return 1/tan(z)
 */
float	cot( float z );

/**
 * Returns both sin and cos for the angle.
 * @param a Angle in radians.
 * @param sina Sine of angle a.
 * @param cosa Cosine of angle a.
 */
void	sincos( float a, float* sina, float* cosa );

/** 
 * Returns value stepped to specified edge, so that 0 is returned for values below the edge.
 * @param edge The edge value.
 * @param v The source value.
 * @return 0 if v<edge, otherwise 1.
 */
float	step( float edge, float v );

/** 
 * Returns value smoothly interpolated between two edges.
 * Useful for smooth transitions.
 * @param edge0 The lower edge value.
 * @param edge1 The upper edge value.
 * @param v The source value.
 * @return 0 if v<edge0, 1 if v>edge1, and Hermite interpolated value (0,1) in between edge0 and edge1.
 */
float	smoothstep( float edge0, float edge1, float v );

/** 
 * Limits value to specified range.
 * @param v Source value.
 * @param vmax Upper limit, inclusive.
 * @param vmin Lower limit, inclusive.
 * @return Source value limited between [vmin,vmax].
 */
float	clamp( float v, float vmin, float vmax );

/** 
 * Limits value to range [0,1].
 * @param v Source value.
 * @return Source value limited between [0,1].
 */
float	saturate( float v );

/**
 * Returns distance to the origin.
 * @param v The point on a line.
 * @return Absolute of the value.
 */
float	length( float v );

/**
 * Returns true if the floating point value is in valid range.
 * @param v Value to check.
 * @return true if -FLT_MAX <= v <= FLT_MAX
 */
bool	check( float v );

/**
 * Returns absolute value.
 * Note the function name to avoid clashes since some platforms might define abs(), min() and max() as macros.
 */
float	absv( float v );

/**
 * Returns minimum of two values.
 * Note the function name to avoid clashes since some platforms might define abs(), min() and max() as macros.
 */
float	minv( float a, float b );

/**
 * Returns maximum of two values.
 * Note the function name to avoid clashes since some platforms might define abs(), min() and max() as macros.
 */
float	maxv( float a, float b );

/** 
 * Converts angle from degrees to radians.
 * @param deg Angle in degrees.
 * @return Angle in radians.
 */
float	radians( float deg );

/** 
 * Converts angle from radians to degrees.
 * @param rad Angle in radians.
 * @return Angle in degrees.
 */
float	degrees( float rad );

/**
 * Calculates 1D Gaussian blur filter kernel with specified size.
 * @param s Width of the kernel. Radius of the kernel is (s-1)/2. s must be at least 1.
 * @param o Standard deviation.
 * @param kernel [out] Array of size [s] receiving filter weights
 * @param kernelsize Size of the output buffer. Must be at least s*s.
 */
void	getGaussianBlurKernel1D( int s, float o, float* kernel, size_t kernelsize );

/**
 * Calculates 2D Gaussian blur filter kernel with specified size.
 * @param s Width of the kernel. Radius of the kernel is (s-1)/2. s must be at least 1.
 * @param o Standard deviation.
 * @param kernel [out] Array of size [s*s] receiving filter weights
 * @param kernelsize Size of the output buffer. Must be at least s*s.
 */
void	getGaussianBlurKernel2D( int s, float o, float* kernel, size_t kernelsize );

/*@}*/

#include <slmath/float_util.inl>

SLMATH_END()

#endif 

// This file is part of 'slmath' C++ library. Copyright (C) 2009 Jani Kajala (kajala@gmail.com). See http://sourceforge.net/projects/slmath/
