#ifndef SLMATH_VEC2_H
#define SLMATH_VEC2_H

#include <slmath/float_util.h>

SLMATH_BEGIN()

/**
 * 2-vector.
 * @ingroup slmath
 */
class vec2
{
public:
	/** Constants related to the class. */
	enum Constants
	{
		/** Number of dimensions in this vector. */
		SIZE = 2,
	};

	/** The first component of the vector. */
	float x;

	/** The second component of the vector. */
	float y;

	/** 
	 * Constructs undefined 2-vector. 
	 */
	vec2();

	/** 
	 * Constructs vector with all components set to the same value.
	 * @param v Each of the vector components will be set to this value.
	 */
	explicit vec2( float v );

	/** 
	 * Constructs the vector from scalars. 
	 * @param x0 The vector X-component (0)
	 * @param y0 The vector Y-component (1)
	 */
	vec2( float x0, float y0 );

	/** Component wise addition. */
	vec2&		operator+=( const vec2& o );

	/** Component wise subtraction. */
	vec2&		operator-=( const vec2& o );

	/** Component wise multiplication. */
	vec2&		operator*=( const vec2& o );

	/** Component wise division. */
	vec2&		operator/=( const vec2& o );

	/** Component wise scalar multiplication. */
	vec2&		operator*=( float s );

	/** Component wise scalar division. */
	vec2&		operator/=( float s );

	/** Returns component at specified index. */
	float&		operator[]( size_t i );

	/** Component wise multiplication. */
	vec2		operator*( const vec2& o ) const;

	/** Component wise division. */
	vec2		operator/( const vec2& o ) const;

	/** Component wise addition. */
	vec2		operator+( const vec2& o ) const;

	/** Component wise subtraction. */
	vec2		operator-( const vec2& o ) const;

	/** Component wise subtraction. */
	vec2		operator-() const;

	/** Component wise scalar multiplication. */
	vec2		operator*( float s ) const;

	/** Component wise scalar division. */
	vec2		operator/( float s ) const;

	/** Returns component at specified index. */
	const float& operator[]( size_t i ) const;

	/** Returns true if vectors are bitwise equal. */
	bool		operator==( const vec2& o ) const;

	/** Returns true if vectors are bitwise inequal. */
	bool		operator!=( const vec2& o ) const;
};


/** 
 * Component wise scalar multiplication. 
 * @ingroup vec_util
 */
vec2			operator*( float s, const vec2& v );

/** 
 * Returns length of the vector. 
 * @ingroup vec_util
 */
float			length( const vec2& v );

/** 
 * Returns dot product of two vectors. 
 * @param a The first vector
 * @param b The second vector
 * @return a dot b
 * @ingroup vec_util
 */
float			dot( const vec2& a, const vec2& b );

/** 
 * Returns the vector normalized to unit (1) length.
 * @ingroup vec_util
 */
vec2			normalize( const vec2& v );

/** 
 * Returns true if all components of the vector are valid numbers. 
 * @ingroup vec_util
 */
bool			check( const vec2& v );

#include <slmath/vec2.inl>

SLMATH_END()

#endif

// This file is part of 'slmath' C++ library. Copyright (C) 2009 Jani Kajala (kajala@gmail.com). See http://sourceforge.net/projects/slmath/
