#ifndef SLMATH_VEC3_H
#define SLMATH_VEC3_H

#include <slmath/vec2.h>

SLMATH_BEGIN()

/**
 * 3-vector.
 * @ingroup slmath
 */
class vec3
{
public:
	/** Constants related to the class. */
	enum Constants
	{
		/** Number of dimensions in this vector. */
		SIZE = 3,
	};

	/** The first component of the vector. */
	float x;

	/** The second component of the vector. */
	float y;

	/** The third component of the vector. */
	float z;

	/** 
	 * Constructs undefined 3-vector. 
	 */
	vec3();

	/** 
	 * Constructs vector with all components set to the same value.
	 * @param v Each of the vector components will be set to this value.
	 */
	explicit vec3( float v );

	/** 
	 * Constructs the vector from scalars. 
	 * @param x0 The vector X-component (0)
	 * @param y0 The vector Y-component (1)
	 * @param z0 The vector Z-component (2)
	 */
	vec3( float x0, float y0, float z0 );

	/** 
	 * Constructs the vector from 2-vector and scalar. 
	 * @param v0 The vector XY-components (0-1)
	 * @param w0 The vector Z-component (2)
	 */
	vec3( const vec2& v0, float w0 );

	/** Component wise addition. */
	vec3&		operator+=( const vec3& o );

	/** Component wise subtraction. */
	vec3&		operator-=( const vec3& o );

	/** Component wise scalar multiplication. */
	vec3&		operator*=( float s );

	/** Component wise scalar division. */
	vec3&		operator/=( float s );

	/** Component wise multiplication. */
	vec3&		operator*=( const vec3& o );

	/** Component wise division. */
	vec3&		operator/=( const vec3& o );

	/** Returns component at specified index. */
	float&		operator[]( size_t i );

	/** Returns sub-vector. */
	vec2&		xy();

	/** Component wise multiplication. */
	vec3		operator*( const vec3& o ) const;

	/** Component wise division. */
	vec3		operator/( const vec3& o ) const;

	/** Component wise addition. */
	vec3		operator+( const vec3& o ) const;

	/** Component wise subtraction. */
	vec3		operator-( const vec3& o ) const;

	/** Component wise subtraction. */
	vec3		operator-() const;

	/** Component wise scalar multiplication. */
	vec3		operator*( float s ) const;

	/** Component wise scalar division. */
	vec3		operator/( float s ) const;

	/** Returns component at specified index. */
	const float& operator[]( size_t i ) const;

	/** Returns true if vectors are bitwise equal. */
	bool		operator==( const vec3& o ) const;

	/** Returns true if vectors are bitwise inequal. */
	bool		operator!=( const vec3& o ) const;

	/** Returns sub-vector. */
	const vec2&	xy() const;
};


/** 
 * Component wise scalar multiplication. 
 * @ingroup vec_util
 */
vec3			operator*( float s, const vec3& v );

/** 
 * Returns length of the vector. 
 * @ingroup vec_util
 */
float			length( const vec3& v );

/** 
 * Returns dot product of two vectors. 
 * @ingroup vec_util
 */
float			dot( const vec3& a, const vec3& b );

/** 
 * Returns the vector normalized. 
 * @ingroup vec_util
 */
vec3			normalize( const vec3& v );

/** 
 * Returns cross product of two vectors. 
 * @ingroup vec_util
 */
vec3			cross( const vec3& a, const vec3& b );

/** 
 * Returns true if all components of the vector are valid numbers. 
 * @ingroup vec_util
 */
bool			check( const vec3& v );

/**
 * Calculates triangle face normal when triangle is defined by counter-clock-wise points.
 */
vec3			facenormal_ccw( const vec3& v0, const vec3& v1, const vec3& v2 );

/**
 * Calculates triangle face normal when triangle is defined by clock-wise points.
 */
vec3			facenormal_cw( const vec3& v0, const vec3& v1, const vec3& v2 );

#include <slmath/vec3.inl>

SLMATH_END()

#endif // SLMATH_VEC3_H

// This file is part of 'slmath' C++ library. Copyright (C) 2009 Jani Kajala (kajala@gmail.com). See http://sourceforge.net/projects/slmath/
