#ifndef SLMATH_VEC4_H
#define SLMATH_VEC4_H

#include <slmath/vec3.h>

SLMATH_BEGIN()

/**
 * 4-vector.
 * @ingroup slmath
 */
SLMATH_ALIGN16 class vec4
{
public:
	/** Constants related to the class. */
	enum Constants
	{
		/** Number of dimensions in this vector. */
		SIZE = 4,
	};

	/** The first component of the vector. */
	float x;

	/** The second component of the vector. */
	float y;

	/** The third component of the vector. */
	float z;

	/** The fourth component of the vector. */
	float w;

	/** Constructs undefined 4-vector. */
	vec4();

	/** 
	 * Constructs vector with all components set to the same value.
	 * @param v Each of the vector components will be set to this value.
	 */
	explicit vec4( float v );

	/** 
	 * Constructs the vector from scalars. 
	 * @param x0 The vector X-component (0)
	 * @param y0 The vector Y-component (1)
	 * @param z0 The vector Z-component (2)
	 * @param w0 The vector W-component (3)
	 */
	vec4( float x0, float y0, float z0, float w0 );

	/** 
	 * Constructs the vector from 2-vector and 2 scalars. 
	 * @param v0 The vector XY-components (0-2)
	 * @param z0 The vector Z-component
	 * @param w0 The vector W-component
	 */
	vec4( const vec2& v0, float z0, float w0 );

	/** 
	 * Constructs the vector from 3-vector and scalar. 
	 * @param v0 The vector XYZ-components (0-3)
	 * @param w0 The vector W-component (3)
	 */
	vec4( const vec3& v0, float w0 );

	/**
	 * Constructs vector from 128bit scalar type.
	 */
	vec4( const SLMATH_M128& xyzw )		{m128()=xyzw;}

	/** Component wise addition. */
	vec4&		operator+=( const vec4& o );

	/** Component wise subtraction. */
	vec4&		operator-=( const vec4& o );

	/** Component wise scalar multiplication. */
	vec4&		operator*=( float s );

	/** Component wise scalar division. */
	vec4&		operator/=( float s );

	/** Component wise multiplication. */
	vec4&		operator*=( const vec4& o );

	/** Component wise division. */
	vec4&		operator/=( const vec4& o );

	/** Returns component at specified index. */
	float&		operator[]( size_t i );

	/** Returns sub-vector. */
	vec3&		xyz();

	/** Returns 128bit scalar type. Internal use only. */
	SLMATH_M128& m128()			{return *(SLMATH_M128*)&x;}

	/** Component wise multiplication. */
	vec4		operator*( const vec4& o ) const;

	/** Component wise division. */
	vec4		operator/( const vec4& o ) const;

	/** Component wise addition. */
	vec4		operator+( const vec4& o ) const;

	/** Component wise subtraction. */
	vec4		operator-( const vec4& o ) const;

	/** Component wise subtraction. */
	vec4		operator-() const;

	/** Component wise scalar multiplication. */
	vec4		operator*( float s ) const;

	/** Component wise scalar division. */
	vec4		operator/( float s ) const;

	/** Returns component at specified index. */
	const float& operator[]( size_t i ) const;

	/** Returns true if vectors are bitwise equal. */
	bool		operator==( const vec4& o ) const;

	/** Returns true if vectors are bitwise inequal. */
	bool		operator!=( const vec4& o ) const;

	/** Returns sub-vector. */
	const vec3&	xyz() const;

	/** Returns 128bit scalar type. Internal use only. */
	const SLMATH_M128& m128() const		{return *(const SLMATH_M128*)&x;}
};


/** 
 * Component wise scalar multiplication. 
 * @ingroup vec_util
 */
vec4			operator*( float s, const vec4& v );

/** 
 * Returns length of the vector. 
 * @ingroup vec_util
 */
float			length( const vec4& v );

/** 
 * Returns dot product of two vectors. 
 * @param a The first vector
 * @param b The second vector
 * @return a dot b
 * @ingroup vec_util
 */
float			dot( const vec4& a, const vec4& b );

/** 
 * Returns the vector normalized to unit (1) length.
 * @ingroup vec_util
 */
vec4			normalize( const vec4& v );

/** 
 * Returns true if all components of the vector are valid numbers. 
 * @ingroup vec_util
 */
bool			check( const vec4& v );

#include <slmath/vec4.inl>

SLMATH_END()

#endif

// This file is part of 'slmath' C++ library. Copyright (C) 2009 Jani Kajala (kajala@gmail.com). See http://sourceforge.net/projects/slmath/
