#ifndef SLMATH_VEC_TEMPL_FUNC_H
#define SLMATH_VEC_TEMPL_FUNC_H

#include <slmath/slmath_pp.h>

SLMATH_BEGIN()

/**
 * \defgroup mat_util Matrix helper functions.
 * @ingroup slmath
 */

/**
 * \defgroup vec_util Vector helper functions.
 * @ingroup slmath
 */

/** 
 * Converts angle from degrees to radians.
 * @param deg Angle in degrees.
 * @return Angle in radians.
 * @ingroup vec_util
 */
template <class T> inline T radians( const T& deg )
{
	return deg * 0.017453292519943295769236907684886f; // pi/180
}

/** 
 * Converts angle from radians to degrees.
 * @param rad Angle in radians.
 * @return Angle in degrees.
 * @ingroup vec_util
 */
template <class T> inline T degrees( const T& rad )
{
	return rad * 57.295779513082320876798154814105f; // 180/pi;
}

/** 
 * Returns linear blend between two values. Formula is x*(1-a)+y*a.
 * @param x The first value.
 * @param y The second value.
 * @param a The blend factor.
 * @return x*(1-a)+y*a
 * @ingroup vec_util
 */
template <class T> T mix( const T& x, const T& y, float a )
{
	return x + (y-x)*a;
}

/** 
 * Returns distance between two vectors. 
 * @param p0 The first point vector.
 * @param p1 The second point vector.
 * @return Distance between p0 and p1, i.e. length(p1-p0).
 * @ingroup vec_util
 */
template <class T> float distance( const T& p0, const T& p1 )
{
	return length( p1 - p0 );
}

/** 
 * Checks if vector points to specified direction.
 * @param n Vector to be reflected if input does not face against reference vector.
 * @param i Input vector.
 * @param nref Normal reference vector.
 * @return If dot(nref<i) < 0 return n, otherwise return -n;
 * @ingroup vec_util
 */
template <class T> T faceforward( const T& n, const T& i, const T& nref )
{
	return dot(nref,i) < 0.f ? n : -n;
}

/** 
 * Reflects vector against specified normal.
 * @param i Input vector.
 * @param n Normal vector. Must be pre-normalized.
 * @return i - 2*dot(n,i)*n
 * @ingroup vec_util
 */
template <class T> T reflect( const T& i, const T& n )
{
	return i - n*( 2.f*dot(n,i) );
}

/** 
 * Refracts vector against specified normal using specified ratio of refraction indices.
 * The result is computed by
 * ndoti = dot(n,i);
 * k = 1.0f - eta*eta*(1.0f - ndoti*ndoti);
 * if k < 0 then return 0 else return eta*i-(eta*ndoti+sqrtf(k))*n
 * @param i Input vector. Must be pre-normalized.
 * @param n Normal vector. Must be pre-normalized.
 * @param eta Ratio of two materials refraction indices.
 * @return Refracted vector.
 * @ingroup vec_util
 */
template <class T> T refract( const T& i, const T& n, float eta )
{
	float ndoti = dot(n,i);
	float k = 1.0f - eta*eta*(1.0f - ndoti*ndoti);
	if ( k < 0.f )
		return T(0.f);
	else
		return i*eta - n*(eta*ndoti+sqrtf(k));
}

/**
 * Returns true if all the values are in valid range.
 * Calls check(v) for each value.
 * @param v Values to check.
 * @param n Number of values to check.
 * @return true if none of v is NaN value.
 * @ingroup vec_util
 */
template <class T> bool check( const T* v, size_t n )
{
	for ( const T* v1 = v+n ; v != v1 ; ++v )
		if ( !check(*v) )
			return false;
	return true;
}

/**
 * Returns vector with components clamped between [minv,maxv].
 * Assumes that class T has SIZE constant defined as number of components.
 * @ingroup vec_util
 */
template <class T> T clamp( const T& v, const T& minv, const T& maxv )
{
	T v1;
	for ( size_t i = 0 ; i < T::SIZE ; ++i )
		v1[i] = v[i] > maxv[i] ? maxv[i] : (v[i] < minv[i] ? minv[i] : v[i]);
	return v1;
}

/**
 * Returns vector with components clamped between [0,1].
 * Assumes that class T has SIZE constant defined as number of components.
 * @ingroup vec_util
 */
template <class T> T saturate( const T& v )
{
	assert( check(v) );
	T v1;
	for ( size_t i = 0 ; i < T::SIZE ; ++i )
		v1[i] = v[i] > 1.f ? 1.f : (v[i] < 0.f ? 0.f : v[i]);
	return v1;
}

/**
 * Returns absolute value component wise.
 * Note the function name to avoid clashes since some platforms might define abs(), min() and max() as macros.
 */
template <class T> T absv( const T& v )
{
	assert( check(v) );
	T v1;
	for ( size_t i = 0 ; i < T::SIZE ; ++i )
		v1[i] = v[i] < 0.f ? -v[i] : v[i];
	return v1;
}

/**
 * Returns minimum of two vectors component wise.
 * Note the function name to avoid clashes since some platforms might define abs(), min() and max() as macros.
 */
template <class T> T minv( const T& a, const T& b )
{
	assert( check(a) );
	assert( check(b) );
	T v1;
	for ( size_t i = 0 ; i < T::SIZE ; ++i )
		v1[i] = a[i] < b[i] ? a[i] : b[i];
	return v1;
}

/**
 * Returns maximum of two vectors component wise.
 * Note the function name to avoid clashes since some platforms might define abs(), min() and max() as macros.
 */
template <class T> T maxv( const T& a, const T& b )
{
	assert( check(a) );
	assert( check(b) );
	T v1;
	for ( size_t i = 0 ; i < T::SIZE ; ++i )
		v1[i] = a[i] < b[i] ? b[i] : a[i];
	return v1;
}

SLMATH_END()

#endif

// This file is part of 'slmath' C++ library. Copyright (C) 2009 Jani Kajala (kajala@gmail.com). See http://sourceforge.net/projects/slmath/
